// Program Code: Coquiduino Braille Printer //
    
// Written by: Jorge Suarez de Freitas + ChatGPT //

#include <queue>
#include <AccelStepper.h>

// Define pins for Motor 1 (Linear Axis with Homing)
#define STEP_PIN1 4
#define DIR_PIN1 3

// Define pins for Motor 2 (Paper Feeder)
#define STEP_PIN2 2
#define DIR_PIN2 1

// Shared enable pin
#define ENABLE_PIN 0
#define LIMIT_SWITCH_PIN 10
#define SOLENOID_PIN 9

// Motor configuration
AccelStepper stepper1(AccelStepper::DRIVER, STEP_PIN1, DIR_PIN1);
AccelStepper stepper2(AccelStepper::DRIVER, STEP_PIN2, DIR_PIN2);

// Variables for homing
bool homingDoneX = false;
bool homingDoneY = false;

// Configuration for paper feeder advancement
const float rollerDiameter = 8.0; // Roller diameter in mm
const float rollerPerimeter = 3.14159 * rollerDiameter; // Roller perimeter
const float stepsPerRevolution = 200.0;
const float microsteps = 16.0;
const float distancePerMicrostep = rollerPerimeter / (stepsPerRevolution * microsteps);

// Configuration for linear axis advancement
const float distancePerMicrostepX = 40.0 / (stepsPerRevolution * microsteps); // 40 mm per revolution with 20 teeth and GT2 (2 mm per tooth)
const float maxXPosition = 184.8; // Maximum position for X-axis in mm
const float maxYPosition = 1000.0; // Maximum position for X-axis in mm

// Define a struct to hold move commands
struct MoveCommand {
  char axis;
  float position;
  bool solenoid;
};

// Queue to hold move commands
std::queue<MoveCommand> commandQueue;

void setup() {
  // Initialize the limit switch and solenoid
  pinMode(LIMIT_SWITCH_PIN, INPUT);
  pinMode(SOLENOID_PIN, OUTPUT);

  // Initialize motors
  stepper1.setMaxSpeed(1000);
  stepper1.setAcceleration(20000);
  stepper2.setMaxSpeed(500);
  stepper2.setAcceleration(10000);

  // Enable motors
  pinMode(ENABLE_PIN, OUTPUT);
  digitalWrite(ENABLE_PIN, LOW);
  delay(100);
  Serial.begin(115200); // Set baud rate to match the one used in the Arduino IDE
  Serial.println("System ready. Sending to home...");
}

void loop() {
  // Read G-code commands from serial port
  if (Serial.available()) {
    String command = Serial.readStringUntil('\n');
    processGCode(command);    
  }

  // Process the command queue
  processCommandQueue();
}

// Function to perform homing of both axes
void homeAllAxes() {
  homeLinearAxis();
  while (!homingDoneX) {
    delay(10); // Wait until homing of the linear axis is complete
  }
  MoveXAxis0();
  homePaperFeeder();

}

void MoveXAxis0() {
  while (stepper1.distanceToGo() != 0) {
    stepper1.run();
  }
  stepper1.moveTo(0); // Move to position 0
  while (stepper1.distanceToGo() != 0) {
    stepper1.run();
  }
}

// Function to perform homing of the linear axis
void homeLinearAxis() {
  stepper1.setMaxSpeed(8000); // Set maximum speed
  stepper1.setSpeed(10000); // Set custom homing speed
  while (digitalRead(LIMIT_SWITCH_PIN) == LOW) { // Adjusted limit switch condition
    stepper1.moveTo(stepper1.currentPosition() + 20); // Move in the positive direction
    stepper1.run();
  }
  stepper1.setCurrentPosition(maxXPosition / distancePerMicrostepX); // Set position to 200 mm
  homingDoneX = true;
  Serial.println("X-Axis homing completed.");
}

// Function to perform homing of the paper feeder
void homePaperFeeder() {
  stepper2.setMaxSpeed(4000); // Set maximum speed
  stepper2.setSpeed(3000); // Set custom homing speed
  while (digitalRead(LIMIT_SWITCH_PIN) == LOW) { // Adjusted limit switch condition
    stepper2.moveTo(stepper2.currentPosition() + 20); // Move in the positive direction
    stepper2.run();
  }
  // Move an additional 10mm in the positive direction
  stepper2.move(10 / distancePerMicrostep); // Convert 10mm to steps
  stepper2.runToPosition(); // Wait for the movement to complete
  delay(200);
   stepper2.move(20 / distancePerMicrostep); // Convert 10mm to steps
  stepper2.runToPosition(); // Wait for the movement to complete
  
  // Set position to 0 mm
  stepper2.setCurrentPosition(0);
  
  homingDoneY = true;
  Serial.println("Paper feeder homing completed.");
}


// Function to process G-code commands
void processGCode(String command) {
  if (command.startsWith("G28")) {
    // Command to perform homing
    homeAllAxes();
  } else if (command.startsWith("G1")) {
    // Move linear axis
    int indexX = command.indexOf('X');
    int indexY = command.indexOf('Y');

    if (indexX >= 0) {
      float posX = command.substring(indexX + 1).toFloat();
      // Check if movement is within safe bounds for X-axis
      if (posX >= 0 && posX <= maxXPosition) {
        commandQueue.push({'X', posX, false});
      } else {
        Serial.println("Error: Movement out of bounds for X-axis.");
      }
    }

    // Move paper feeder
    if (indexY >= 0) {
      float posY = command.substring(indexY + 1).toFloat();
      // Check if movement is within safe bounds for Y-axis
      if (posY >= 0 && posY <= maxYPosition) {
        commandQueue.push({'Y', posY, false});
      } else {
        Serial.println("Error: Movement out of bounds for Y-axis.");
      }
    }
  } else if (command.startsWith("P")) { 
    commandQueue.push({'P', 0, true});
  } else if (command.startsWith("HELLOWORLD")) {
    HELLOWORLD();
  } else if (command.startsWith("NEIL")) {
    NEIL();
  } else if (command.startsWith("FAB")) {
    THISISFABACADEMY2024();
  }
}



void HELLOWORLD() {
  // Define starting positions
  int startX = 0;
  int startY = 0;


  // Print "H"
  commandQueue.push({'X', startX, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "E" 
  commandQueue.push({'X', startX, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "L"
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "L"
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "O" 
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 15;

  // Print "W" 
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "O"
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "R" 
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "L"
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY + 10, false});
  commandQueue.push({'P', 0, true});

  startX += 10;

  // Print "D" 
  commandQueue.push({'X', startX + 5, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY, false});
  commandQueue.push({'P', 0, true});
  commandQueue.push({'X', startX + 10, false});
  commandQueue.push({'Y', startY + 5, false});
  commandQueue.push({'P', 0, true});

  NEIL();
  
  THISISFABACADEMY2024();



  // Move pen away from drawing area
  commandQueue.push({'Y', 300, false});
  commandQueue.push({'X', 180, false});
  }



void NEIL() {
    // Define starting positions
    int startX = 0;
    int startY = 30;

    // Print "H" 
    commandQueue.push({'X', startX + 0, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 0, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});

    startX += 10;

    // Print "I" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});

    startX += 15;

    // Print "N" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
   
    startX += 10;

    // Print "E" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});

    startX += 10;
    
    // Print "I" (2nd and 3rd dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY , false});
    commandQueue.push({'P', 0, true});

    startX += 10;

    // Print "L" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
    

    


    // // Move pen away from drawing area
    // commandQueue.push({'Y', startY + 300, false});
    // commandQueue.push({'X', startX + 180, false});
    }


void THISISFABACADEMY2024() {
    // Define starting positions
    int startX = 0;
    int startY = 60;

    // Print "T" 
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY , false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
    startX += 10;
    
    // Print "H"
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "I" (2nd and 4th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "S" (1st, 2nd, 4th, and 5th dots)
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print space (move pen to the next character position)
    startX += 10;

    // Print "I" (2nd and 4th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "S" (1st, 2nd, 4th, and 5th dots)
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});

    // Move to the next line
    startX = 0;
    startY += 20;

    // Print "F" (1st, 2nd, 3rd, and 5th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX , false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "A" (1st dot)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "B" (1st and 2nd dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print space (move pen to the next character position)
    startX += 10;

    // Print "A" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "C" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "A" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "D" (1st, 4th, and 5th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "E" (1st and 5th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "M" (1st, 3rd, and 4th dots)
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
    startX += 10;

    // Print "Y" 
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 5, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});
    commandQueue.push({'X', startX + 5, false});
    commandQueue.push({'Y', startY + 10, false});
    commandQueue.push({'P', 0, true});

    // // Move pen away from drawing area
    // commandQueue.push({'Y', 300, false});
    // commandQueue.push({'X', 180, false});
  }



// Function to move the linear axis

void moveLinearAxis(float position) {
  if (homingDoneX) {
    long steps = position / distancePerMicrostepX; // Convert mm to steps
    stepper1.moveTo(steps);
  } else {
    Serial.println("Error: Linear axis homing not completed.");
  }
}

// Function to move the paper feeder
void movePaperFeeder(float distance) {
  if (homingDoneY) {
    long steps = distance / distancePerMicrostep; // Convert mm to steps
    stepper2.moveTo(steps);
  } else {
    Serial.println("Error: Paper feeder homing not completed.");
  }
}

// Function to activate the solenoid
void activateSolenoid() {
  digitalWrite(SOLENOID_PIN, HIGH);
  delay(300);
  digitalWrite(SOLENOID_PIN, LOW);
  delay(100);
}

// Function to process the command queue
void processCommandQueue() {
  if (!commandQueue.empty()) {
    MoveCommand cmd = commandQueue.front();

    if (stepper1.distanceToGo() == 0 && stepper2.distanceToGo() == 0) {
      if (cmd.axis == 'X') {
        moveLinearAxis(cmd.position);
      } else if (cmd.axis == 'Y') {
        movePaperFeeder(cmd.position);
      } else if (cmd.solenoid) {
        activateSolenoid();
      }
      commandQueue.pop();
    }
  }

  // Run steppers
  stepper1.run();
  stepper2.run();
}

